import typing

from frozendict import frozendict
import numpy as np

from matplotlib.axes import Axes as mpl_Axes
from matplotlib.collections import PathCollection as mpl_PathCollection
from matplotlib.container import ErrorbarContainer as mpl_ErrorbarContainer

from ._stub_edge_mark import stub_edge_mark


def stub_all_clipped_values(
    ax: mpl_Axes,
    *,
    marker_kws: typing.Dict = frozendict(),
    offset: float = 0.1,
) -> None:
    """Draw edge markers for points lying outside the plot's bounds.

    The original point and associated error bar (if any) is moved into the
    margin with a note indicating the original location. If the point is out of
    xlim and out of ylim, the point will be retrieved with two edge markers
    (horizontal and vertical).

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object to draw markers on.
    marker_kws : dict, optional
        Keyword arguments forwarded to matplotlib `plot` for edge markers.
    offset : float, default 0.1
        How far outside axis viewport to draw edge markers, proportional to
        axis viewport height or width.

    Notes
    -----
    Full error bar support for values below lower {x,y}lim remains be
    implemented. Current implementation assumes each error bar is associated
    with a scatter point.
    """
    xlim, ylim = ax.get_xlim(), ax.get_ylim()

    # move out of bounds points, if any
    for collection in ax.collections:
        if isinstance(collection, mpl_PathCollection):
            # Extract points from the line
            new_offsets = []
            for x, y in collection.get_offsets():
                # Draw edge marker if the point is out of bounds
                if x < xlim[0] or x > xlim[1] or y < ylim[0] or y > ylim[1]:
                    x, y = stub_edge_mark(
                        ax,
                        x,
                        y,
                        marker_kws=marker_kws,
                        offset=offset,
                    )
                new_offsets.append((x, y))

            collection.set_offsets(new_offsets)
            collection.set(clip_on=False)

    # move out of bounds error bars, if any
    x_width, y_height = np.ptp(xlim), np.ptp(ylim)
    x_thresh, y_thresh = xlim[1], ylim[1]
    x_offset = x_thresh + x_width * offset
    y_offset = y_thresh + y_height * offset
    for container in ax.containers:
        if isinstance(container, mpl_ErrorbarContainer):
            # Unpack the container
            (
                plotline,  # Line2D instance of x, y plot markers and/or line
                caplines,  # A tuple of Line2D instances of the error bar caps
                # A tuple of LineCollection with the horizontal and vertical
                # error ranges.
                barlinecols,
            ) = container

            # Adjust each error bar
            for barlinecol in barlinecols:
                segments = barlinecol.get_segments()
                new_segments = []
                for segment in segments:
                    (x0, y0), (x1, y1) = segment
                    if x0 > x_thresh:
                        sep = x1 - x0
                        x0, x1 = x_offset - sep / 2, x_offset + sep / 2
                    if y0 > y_thresh:
                        sep = y1 - y0
                        y0, y1 = y_offset - sep / 2, y_offset + sep / 2
                    new_segment = [(x0, y0), (x1, y1)]
                    new_segments.append(new_segment)

                barlinecol.set_segments(new_segments)
                barlinecol.set(clip_on=False)



import numpy as np
import pytest

from outset.tweak import TweakReflect


test_data = [
    # Format: (horizontal, vertical, input_vertices, expected_output)
    (
        True,
        False,
        [(1, 1), (2, 2), (3, 3), (4, 4)],
        [(3, 1), (2, 2), (1, 3), (0, 4)],
    ),  # Horizontal reflection
    (
        False,
        True,
        [(1, 1), (2, 2), (3, 3), (4, 4)],
        [(1, 3), (2, 2), (3, 1), (4, 0)],
    ),  # Vertical reflection
    (
        True,
        True,
        [(1, 1), (2, 2), (3, 3), (4, 4)],
        [(3, 3), (2, 2), (1, 1), (0, 0)],
    ),  # Both reflections
    (
        None,
        None,
        [(1, 1), (2, 2), (3, 3), (4, 4)],
        [(3, 1), (2, 2), (1, 3), (0, 4)],
    ),  # Default (horizontal)
]


@pytest.mark.parametrize(
    "horizontal, vertical, input_vertices, expected_output", test_data
)
def test_TweakReflect(horizontal, vertical, input_vertices, expected_output):
    tweak = TweakReflect(horizontal=horizontal, vertical=vertical)
    result = tweak(input_vertices)
    assert np.allclose(result, expected_output)



import itertools as it
import numbers
import typing

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes as mpl_Axes

from ._auxlib.calc_aspect_ import calc_aspect
from ._auxlib.calc_outer_pad_ import calc_outer_pad
from ._auxlib.is_axes_unset_ import is_axes_unset
from ._auxlib.robust_groupby_ import robust_groupby
from ._auxlib.set_aspect_ import set_aspect
from ._draw_marquee import draw_marquee
from .mark._MarkNumericalBadges import MarkNumericalBadges


def marqueeplot(
    data: pd.DataFrame,
    *,
    x: str,
    y: str,
    hue: typing.Optional[str] = None,
    hue_order: typing.Optional[typing.Sequence[str]] = None,
    outset: typing.Optional[str] = None,
    outset_order: typing.Optional[typing.Sequence[str]] = None,
    ax: typing.Optional[mpl_Axes] = None,
    color: typing.Optional[str] = None,
    frame_inner_pad: typing.Union[float, typing.Tuple[float, float]] = 0.1,
    frame_outer_pad: typing.Union[float, typing.Tuple[float, float]] = 0.1,
    frame_outer_pad_unit: typing.Literal["axes", "figure", "inches"] = "axes",
    leader_tweak: typing.Union[
        typing.Callable, typing.Type
    ] = lambda x, *args, **kwargs: x,
    mark_glyph: typing.Union[
        typing.Callable, typing.Type, None
    ] = MarkNumericalBadges,
    palette: typing.Optional[typing.Sequence] = None,
    preserve_aspect: typing.Optional[bool] = False,
    tight_axlim: bool = False,
    **kwargs,
) -> mpl_Axes:
    """Plot marquee annotations to contain subsets of data from a pandas
    DataFrame.

    Provides a seaborn-like axis-level interface for `draw_marquee`. Details on
    marquee annotation structure and configuration can be found in that
    function's docstring.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing the data to be marquee-annotated.
    x : str
        Column name in `data` for x-coordinate values of data positions.
    y : str
        Column name in `data` for y-coordinate values of data positions.
    hue : str, optional
        Column name in `data` for grouping data by color.

        If provided, colors are chosen according to palette.
    hue_order : Sequence[str], optional
        Order for plotting the categorical levels of `hue`.
    outset : str, optional
        Column name in `data` for producing different-colored annotated subsets.
    outset_order : Sequence[str], optional
        Order for plotting the categorical levels of `outset`.
    ax : mpl_axes.Axes, optional
        Matplotlib Axes object to draw the plot on.
    color : str, optional
        Color for all elements in the plot, overriding the `palette`.
    frame_inner_pad : Union[float, Tuple[float, float]], default 0.1
        Padding from data range to rectangular boundary.
    frame_outer_pad : Union[float, Tuple[float, float]], default 0.1
        Padding from frame boundary to axis viewport.
    frame_outer_pad_unit : Literal["axes", "figure", "inches"], default "axes"
        How should outer padding be specified?

        If 'axes' or 'figure', padding is specified as a fraction of the axes
        or figure size, respectively. If 'inches', padding is specified in
        inches.
    leader_tweak : Callable, default identity
        Callable or functor type to modify the callout leader vertices before
        drawing.
    mark_glyph : Union[Callable, Type, None], optional
        Callable or functor type to draw a glyph at the end of the callout.
    palette : Sequence, optional
        Color palette for plotting elements.
    preserve_aspect: bool, default False
        If True, finalizing by applying initial axes aspect. If None, restore initial axes aspect unless axes are unset.
    tight_axlim : bool, default False
        Whether to shrink axes limits to fit data range.
    **kwargs : dict
        Keyword arguments to adjust marquee sizing and styling.

        See `outset.draw_marquee` for available options.

    Returns
    -------
    matplotlib.axes.Axes
        The Matplotlib axes containing the plot with annotated regions.

    See Also
    --------
    outset.OutsetGrid
        Figure-level interface for creating plots with marquee annotations.
    outset.draw_marquee
        Low-level function for drawing marquee annotations.
    """
    if ax is None:
        ax = plt.gca()

    initial_axlim = ax.get_xlim(), ax.get_ylim()
    if preserve_aspect or (preserve_aspect is None and not is_axes_unset(ax)):
        initial_aspect = calc_aspect(ax)
    else:
        initial_aspect = None

    if palette is None:
        palette = sns.color_palette()

    if hue is not None and color is not None:
        raise ValueError(f"cannot specify both hue={hue} and color={color}")

    if isinstance(mark_glyph, type):
        mark_glyph = mark_glyph()

    if isinstance(leader_tweak, type):
        leader_tweak = leader_tweak()

    data = data.copy()

    # assemble data groups
    if hue is None:
        hue = "_dummy_hue"
        assert hue not in data.columns
        data[hue] = 0
        palette = [color]

    if hue_order is None:
        hue_order = sorted(data[hue].unique())

    color_lookup = dict(
        zip(
            hue_order,
            it.cycle(palette) if color is None else it.repeat(color),
        )
    )

    if outset is None:
        outset = "_dummy_outset"
        assert outset not in data.columns
        data[outset] = 0

    if outset_order is None:
        outset_order = sorted(data[outset].unique())

    data = data[
        data[hue].isin(hue_order) & data[outset].isin(outset_order)
    ].copy()

    assert "_dummy_hue_key" not in data.columns
    assert "_dummy_outset_key" not in data.columns
    data["_dummy_hue_key"] = data[hue].map([*hue_order].index)
    data["_dummy_outset_key"] = data[outset].map([*outset_order].index)
    data.sort_values(["_dummy_outset_key", "_dummy_hue_key"], inplace=True)

    # need to solve for and apply outer padding prior to plotting to ensure
    # consistency...
    _prepad_axlim(
        data=data,
        x=x,
        y=y,
        hue=hue,
        outset=outset,
        ax=ax,
        frame_inner_pad=frame_inner_pad,
        frame_outer_pad=frame_outer_pad,
        frame_outer_pad_unit=frame_outer_pad_unit,
        tight_axlim=tight_axlim,
    )

    for (_outset_value, hue_value), subset in data.groupby(
        [outset, hue],
        sort=False,
    ):
        assert len(subset)
        xlim = [subset[x].min(), subset[x].max()]
        ylim = [subset[y].min(), subset[y].max()]
        selected_color = color_lookup[hue_value]
        draw_marquee(
            frame_xlim=xlim,
            frame_ylim=ylim,
            ax=ax,
            color=selected_color,
            frame_inner_pad=frame_inner_pad,
            frame_outer_pad=(0, 0),  # already padded by prepad_axlim...
            leader_tweak=leader_tweak,
            mark_glyph=mark_glyph,
            **kwargs,
        )

    if initial_aspect is not None and not np.allclose(
        np.array(initial_axlim),
        np.array(
            (
                ax.get_xlim(),
                ax.get_ylim(),
            )
        ),
    ):
        set_aspect(ax, initial_aspect)

    return ax


def _prepad_axlim(
    data: pd.DataFrame,
    x: str,
    y: str,
    hue: typing.Optional[str],
    outset: typing.Optional[str],
    frame_inner_pad: typing.Union[float, typing.Tuple[float, float]],
    frame_outer_pad: typing.Union[float, typing.Tuple[float, float]],
    frame_outer_pad_unit: typing.Literal["axes", "figure", "data"],
    tight_axlim: bool,
    ax: typing.Optional[mpl_Axes] = None,
) -> None:
    """Calculate padded frame bounds and, if necessary, grow axes limits to
    include them."""
    if ax is None:
        ax = plt.gca()

    # precalculate frames with inner padding
    framex_values, framey_values = [], []
    for _, subset in robust_groupby(data, by=[outset, hue], sort=False):
        assert len(subset)

        is_number = isinstance(frame_inner_pad, numbers.Number)
        if is_number:
            # convert to absolute units to prevent weird effects from
            # successive calls to draw_marquee
            frame_inner_pad_x, frame_inner_pad_y = (
                frame_inner_pad * (np.ptp(subset[x]) or np.ptp(ax.get_xlim())),
                frame_inner_pad * (np.ptp(subset[y]) or np.ptp(ax.get_ylim())),
            )
        else:
            frame_inner_pad_x, frame_inner_pad_y = frame_inner_pad

        framex_values.extend(
            [
                subset[x].min() - frame_inner_pad_x,
                subset[x].max() + frame_inner_pad_x,
            ],
        )
        framey_values.extend(
            [
                subset[y].min() - frame_inner_pad_y,
                subset[y].max() + frame_inner_pad_y,
            ],
        )

    if is_axes_unset(ax) or tight_axlim:
        if framex_values and np.ptp(framex_values):
            ax.set_xlim(min(framex_values), max(framex_values))
        if framey_values and np.ptp(framey_values):
            ax.set_ylim(min(framey_values), max(framey_values))
    else:
        (x0, x1), (y0, y1) = ax.get_xlim(), ax.get_ylim()
        ax.set_xlim(min(*framex_values, x0), max(*framex_values, x1))
        ax.set_ylim(min(*framey_values, y0), max(*framey_values, y1))

    pad_x, pad_y = calc_outer_pad(ax, frame_outer_pad, frame_outer_pad_unit)
    if len(data):
        lowerx, upperx = (
            np.min(framex_values) - pad_x,
            np.max(framex_values) + pad_x,
        )
        lowery, uppery = (
            np.min(framey_values) - pad_y,
            np.max(framey_values) + pad_y,
        )
    else:
        lowerx, upperx = ax.get_xlim()
        lowery, uppery = ax.get_ylim()

    if tight_axlim or is_axes_unset(ax):
        pass
    else:
        lowerx = min(lowerx, ax.get_xlim()[0])
        lowery = min(lowery, ax.get_ylim()[0])
        upperx = max(upperx, ax.get_xlim()[1])
        uppery = max(uppery, ax.get_ylim()[1])

    # apply axis limit to incorporate outer padding
    ax.set_xlim(lowerx, upperx)
    ax.set_ylim(lowery, uppery)



from outset._auxlib.calc_aspect_ import calc_aspect
from outset._auxlib.equalize_aspect_ import equalize_aspect

from matplotlib import pyplot as plt
import pytest


def test_equalize_aspect():
    # Create a figure with two subplots having different aspect ratios
    fig, (ax1, ax2) = plt.subplots(1, 2)
    ax1.set_xlim(0, 10)
    ax1.set_ylim(0, 20)  # Aspect ratio 2:1
    ax2.set_xlim(0, 10)
    ax2.set_ylim(0, 10)  # Aspect ratio 1:1

    before_aspect1 = calc_aspect(ax1)
    before_aspect2 = calc_aspect(ax2)

    # Apply the equalize_aspect function
    after_aspect = equalize_aspect([ax1, ax2])

    # Calculate aspect ratios after equalization
    after_aspect_ax1 = calc_aspect(ax1)
    after_aspect_ax2 = calc_aspect(ax2)

    # Check if both aspect ratios are equal and match the calculated new aspect ratio
    assert after_aspect_ax1 == pytest.approx(after_aspect)
    assert after_aspect_ax2 == pytest.approx(after_aspect)

    expected_after_aspect = (before_aspect1 * before_aspect2) ** 0.5
    assert after_aspect == pytest.approx(expected_after_aspect)


def test_equalize_aspect_singleton():
    plt.gca().set_xlim(0, 10)
    plt.gca().set_ylim(1, 6)
    aspect_before = calc_aspect(plt.gca())
    assert equalize_aspect([plt.gca()]) == aspect_before
    assert plt.gca().get_xlim() == (0, 10)
    assert plt.gca().get_ylim() == (1, 6)


def test_equalize_aspect_empty():
    assert equalize_aspect([]) == 1.0



import matplotlib.pyplot as plt

from outset._auxlib.is_axes_unset_ import is_axes_unset


def test_empty_axis():
    fig, ax = plt.subplots()
    assert is_axes_unset(ax) is True


def test_empty_axis_gca():
    assert is_axes_unset(plt.gca()) is True


def test_axis_with_children():
    fig, ax = plt.subplots()
    ax.plot([0, 1], [0, 1])  # Add a line plot to the axis
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    assert is_axes_unset(ax) is False


def test_manually_set_limits():
    fig, ax = plt.subplots()
    ax.set_xlim(0.0, 2.0)
    assert is_axes_unset(ax) is False


def test_auto_set_limits():
    fig, ax = plt.subplots()
    ax.plot([0, 1], [0, 2])  # Add a line plot to the axis
    assert is_axes_unset(ax) is False



import typing

from matplotlib import axes as mpl_axes

_color_t = typing.Union[typing.Tuple, str]

from ._MarkAlphabeticalBadges import MarkAlphabeticalBadges


class MarkRomanBadges:
    """Functor to mark roman numeral badges numbered by itertools-like count.

    This class renders markers as roman numerals overlaid onto a circular badge
    with a slightly larger circular underlay. Both lower and uppercase numerals
    are supported.

    Notes
    -----
    Due to unicode limitations values greater than twelve will not display
    correctly. Numeral representation for zero is not supported.
    """

    _ftor: MarkAlphabeticalBadges

    def __init__(
        self: "MarkRomanBadges",
        start: int = 1,
        step: int = 1,
        *,
        upper: bool = False,
        **kwargs,
    ) -> None:
        """Initialize functor.

        Parameters
        ----------
        start : int, default 1
            The starting number for the numerical badges.

            Must be between 1 and 12, inclusive.
        step : int, default 1
            The step size for the numerical badges.
        upper : bool, default False
            Should numerals be uppercase?
        kwwargs : dict
            Additional kwargs forward to `__call__`.
        """
        if start < 1 or start > 12:
            raise ValueError(f"Start value {start} outside supported range.")
        # RE unicode roman numerals, see
        # https://www.johndcook.com/blog/2020/10/07/roman-numerals/ and
        # https://en.wikipedia.org/wiki/Numerals_in_Unicode#Roman_numerals
        base = 0x215F if upper else 0x216F
        self._ftor = MarkAlphabeticalBadges(
            start=chr(base + start), step=step, **kwargs
        )

    ftor = MarkAlphabeticalBadges()

    def __call__(
        self: "MarkRomanBadges",
        x: float,
        y: float,
        ax: typing.Optional[mpl_axes.Axes] = None,
        *,
        color: str = "black",
        color_accent: typing.Optional[str] = None,
        color_numeral: _color_t = "color_accent",
        color_badge: _color_t = "color",
        color_underlay: _color_t = "color_accent",
        marker_badge: str = "o",
        marker_underlay: str = "o",
        markersize: float = 22,
        numeral_edgewidth: float = 0,
        scale_numeral: float = 0.4,
        scale_badge: float = 0.8,
        **kwargs,
    ) -> None:
        """Draw roman numeral badge marker at specified location.

        Numeral values increment with each call to this functor.

        Parameters
        ----------
        x : float
            The x-coordinate where the numerical badge will be placed.
        y : float
            The y-coordinate where the numerical badge will be placed.
        ax : mpl_axes.Axes, optional
            The axes object on which to draw the marker. If None, plt.gca()
            will be used.
        color : str, default "black"
            The primary color for the glyph.
        color_accent : Optional[str], optional
            The default accent color for the glyph.

            If `color` is "white", defaults "black". Otherwise, defaults
            "white".
        color_numeral : Union[str, tuple], default "white"
            The color of the numeral.
        color_badge : Union[str, tuple], default "color"
            The color of the badge.

            If "color", primary color will be substituted.
        color_underlay : Union[str, tuple], default "color_accent"
            The color of the underlay.

            If "color_accent", accent color will be substituted.
        marker_badge : str, default "o"
            The marker style for the badge.
        marker_underlay : str, default "o"
            The marker style for the underlay.
        markersize : float, default 22
            Size for glyph's largest marker element.
        numeral_edgewidth : float, default 1
            The edge width of the numeral marker.
        scale_numeral : float, default 0.4
            The scaling factor for the numeral size.
        scale_badge : float, default 0.8
            The scaling factor for the badge size.
        **kwargs : dict, optional
            Additional keyword arguments forward to matplotlib `plot`.

        Returns
        -------
        None
        """
        self._ftor(
            x=x,
            y=y,
            ax=ax,
            color=color,
            color_accent=color_accent,
            color_letter=color_numeral,
            color_badge=color_badge,
            color_underlay=color_underlay,
            letter_edgewidth=numeral_edgewidth,
            marker_badge=marker_badge,
            marker_underlay=marker_underlay,
            markersize=markersize,
            scale_letter=scale_numeral,
            scale_badge=scale_badge,
            **{
                "linecolor": "none",
                **kwargs,
            },
        )



import typing

from matplotlib import axes as mpl_axes

from ._MarkInlaidAsterisk import mark_inlaid_asterisk

_color_t = typing.Union[typing.Tuple, str]


class MarkNumericalBadges:
    """Functor to mark badges numbered by itertools-like count.

    This class renders markers as numbers overlaid onto a circular badge with a
    slightly larger circular underlay.
    """

    _counter: int
    _step: int
    _kws: dict

    def __init__(
        self: "MarkNumericalBadges", start: int = 1, step: int = 1, **kwargs
    ) -> None:
        """Initialize functor.

        Parameters
        ----------
        start : int, default 1
            The starting number for the numerical badges.
        step : int, default 1
            The step size for the numerical badges.
        kwargs : dict
            Additional kwargs will forward to `__call_`.
        """
        self._counter = start
        self._step = step
        self._kws = kwargs

    def __call__(
        self: "MarkNumericalBadges",
        x: float,
        y: float,
        ax: typing.Optional[mpl_axes.Axes] = None,
        *,
        color: str = "black",
        color_accent: typing.Optional[str] = None,
        color_numeral: _color_t = "color_accent",
        color_badge: _color_t = "color",
        color_underlay: _color_t = "color_accent",
        marker_badge: str = "o",
        marker_underlay: str = "o",
        markersize: float = 22,
        numeral_edgewidth: float = 1,
        scale_numeral: float = 0.4,
        scale_badge: float = 0.8,
        **kwargs,
    ) -> None:
        """Draw numbered badge marker at specified location.

        Numeral values increment with each call to this functor.

        Parameters
        ----------
        x : float
            The x-coordinate where the numerical badge will be placed.
        y : float
            The y-coordinate where the numerical badge will be placed.
        ax : mpl_axes.Axes, optional
            The axes object on which to draw the marker. If None, plt.gca()
            will be used.
        color : str, default "black"
            The primary color for the glyph.
        color_accent : Optional[str], optional
            The default accent color for the glyph.

            If `color` is "white", defaults "black". Otherwise, defaults
            "white".
        color_numeral : Union[str, tuple], default "white"
            The color of the numeral.
        color_badge : Union[str, tuple], default "color"
            The color of the badge.

            If "color", primary color will be substituted.
        color_underlay : Union[str, tuple], default "color_accent"
            The color of the underlay.

            If "color_accent", accent color will be substituted.
        marker_badge : str, default "o"
            The marker style for the badge.
        marker_underlay : str, default "o"
            The marker style for the underlay.
        markersize : float, default 22
            Size for glyph's largest marker element.
        numeral_edgewidth : float, default 1
            The edge width of the numeral marker.
        scale_numeral : float, default 0.4
            The scaling factor for the numeral size.
        scale_badge : float, default 0.8
            The scaling factor for the badge size.
        **kwargs : dict, optional
            Additional keyword arguments forward to matplotlib `plot`.

        Returns
        -------
        None
        """
        mark_inlaid_asterisk(
            **{
                "linecolor": "none",
                **self._kws,
                **dict(
                    x=x,
                    y=y,
                    ax=ax,
                    asterisk_edgewidth=numeral_edgewidth,
                    color=color,
                    color_accent=color_accent,
                    color_asterisk_edge=color_numeral,
                    color_asterisk_face=color_numeral,
                    color_badge=color_badge,
                    color_underlay=color_underlay,
                    marker=f"${self._counter}$",
                    marker_badge=marker_badge,
                    marker_underlay=marker_underlay,
                    markersize=markersize,
                    scale_asterisk=scale_numeral,
                    scale_badge=scale_badge,
                ),
                **kwargs,
            },
        )
        self._counter += self._step



import numpy as np

from outset.stub import CalcBoundsIQR


def test_empty_dataset():
    data = np.array([])
    lower_bound_empty, upper_bound_empty = CalcBoundsIQR()(data)
    assert lower_bound_empty == 0.0
    assert upper_bound_empty == 0.0


def test_singleton_dataset():
    data = np.array([3])
    lower_bound_same, upper_bound_same = CalcBoundsIQR()(data)
    assert lower_bound_same == 3.0
    assert upper_bound_same == 3.0


def test_same_value_dataset():
    data_same_value = np.array([3] * 10)
    lower_bound_same, upper_bound_same = CalcBoundsIQR()(data_same_value)
    assert lower_bound_same == 3.0
    assert upper_bound_same == 3.0


def test_typical_dataset():
    data = np.array([1, 2, 3, 4, 5, 100])
    assert CalcBoundsIQR(iqr_multiplier=2.0)(data) == (-2.75, 9.75)


def test_zero_multiplier():
    data = np.array([1, 2, 3, 4, 5, 100])
    assert CalcBoundsIQR(iqr_multiplier=0.0)(data) == (2.25, 4.75)



"""Utility functions."""

from ._layout_corner_insets import layout_corner_insets
from ._NamedFrames import NamedFrames
from ._SplitKwarg import SplitKwarg

__all__ = [
    "layout_corner_insets",
    "NamedFrames",
    "SplitKwarg",
]



from outset.util import SplitKwarg


def test_split_kwarg_initialization():
    source_val = "source_data"
    outset_val = "outset_data"

    split_kwarg = SplitKwarg(source=source_val, outset=outset_val)

    assert split_kwarg.source == source_val
    assert split_kwarg.outset == outset_val



import random

from keras.datasets import mnist
from matplotlib import pyplot as plt
from matplotlib.patheffects import withStroke
import numpy as np
from numpy import reshape
import pandas as pd
import pytest
from sklearn.manifold import TSNE
from sklearn.datasets import load_iris
import seaborn as sns

import outset


random.seed(4)
np.random.seed(4)


# adapted from https://stackoverflow.com/a/64554001
def jitter(values):
    return values + np.random.normal(0, 5, values.shape)


# adapted from https://www.datatechnotes.com/2020/11/tsne-visualization-example-in-python.html
@pytest.mark.integration
def test_make_logo():
    (x_train, y_train), (_, _) = mnist.load_data()
    x_train = x_train[:3000]
    y_train = y_train[:3000]
    x_mnist = reshape(
        x_train, [x_train.shape[0], x_train.shape[1] * x_train.shape[2]]
    )

    tsne = TSNE(n_components=2, verbose=1, random_state=123)
    z = tsne.fit_transform(x_mnist)
    df = pd.DataFrame()
    df["y"] = y_train
    df["comp-1"] = jitter(z[:, 0])
    df["comp-2"] = jitter(z[:, 1])

    xlim = (-60, 80)
    ylim = (-60, 60)
    og = outset.OutsetGrid(
        data=[
            ((25, -3), (33, 5)),
            ((-2.5, -26), (20, -18)),
        ],
        aspect=1.2,
        marqueeplot_kws=dict(frame_edge_kws={"linewidth": 2}),
        marqueeplot_source_kws=dict(
            frame_face_kws={"alpha": 0.2},
            leader_face_kws={"alpha": 0.0, "linestyle": (0, (1, 0.5))},
            leader_stretch=0.12,
            mark_retract=0.25,
        ),
        zorder=4,
    )
    og.broadcast_source(
        sns.scatterplot,
        data=df,
        x="comp-1",
        y="comp-2",
        hue=df.y.tolist(),
        palette=sns.color_palette("hls", 10),
        alpha=0.6,
        legend=False,
        s=30,
        ax=og.source_axes,
    )
    og.marqueeplot()
    og.broadcast_outset(
        sns.scatterplot,
        data=df,
        x="comp-1",
        y="comp-2",
        hue=df.y.tolist(),
        palette=sns.color_palette("hls", 10),
        alpha=0.6,
        legend=False,
        s=60,
    )
    og.broadcast(sns.despine)

    # Add text to the plot with outline
    for color, coord, rot, txt in (
        (
            "lightblue",
            (xlim[0] + np.ptp(xlim) * 0.3, ylim[0] + np.ptp(ylim) * 0.55),
            -20,
            "outset",
        ),
        (
            "mediumaquamarine",
            (xlim[0] + np.ptp(xlim) * 0.65, ylim[0] + np.ptp(ylim) * 0.5),
            0,
            "out",
        ),
        (
            "mediumpurple",
            (xlim[0] + np.ptp(xlim) * 0.4, ylim[0] + np.ptp(ylim) * 0.3),
            0,
            "set",
        ),
    ):
        og.broadcast_outset(
            plt.text,
            *coord,
            txt,
            clip_on=True,
            fontsize=48,
            fontstyle="italic",
            color=color,
            rotation=rot,
            path_effects=[withStroke(linewidth=20, foreground="white")],
            zorder=5,
        )
        og.source_axes.text(
            *coord,
            txt,
            fontsize=8,
            fontstyle="italic",
            color=color,
            rotation=rot,
            path_effects=[withStroke(linewidth=5, foreground="white")],
            zorder=3,
        )

    # Strip Axis Text and Labels
    for ax in og.axes.flat:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_title("")

    # Adjust Spacing
    og.fig.tight_layout(w_pad=1, h_pad=1)

    outpath = "/tmp/outset_logo.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")
